@@ -10,10 +10,26 @@ from .ADQLParser import ADQLParser
 from .ADQLParserVisitor import ADQLParserVisitor
 from .ADQLParserListener import ADQLParserListener
 
+from ..exceptions import QueryError, QuerySyntaxError
+
+
+# Function names need to be trecognized because there whitespace 
+# between the name and left parenthesis is not allowed and needs to be
+# deleted.
+adql_function_names = ('ABS', 'CEILING', 'DEGREES', 'EXP', 'FLOOR', 'LOG',
+                       'LOG10', 'MOD', 'PI', 'POWER', 'RADIANS', 'SQRT',
+                       'TRUNCATE', 'COUNT', 'ACOS', 'ASIN', 'ATAN', 'ATAN2',
+                       'COS', 'COT', 'SIN', 'TAN')
+
 
 def _remove_children(ctx):
-        for i in range(ctx.getChildCount() - 1):
-            ctx.removeLastChild()
+    for i in range(ctx.getChildCount() - 1):
+        ctx.removeLastChild()
+
+
+def _process_regular_identifier(ctx_text):
+        ri = ctx_text.rstrip("'").lstrip("'").rstrip('"').lstrip('"')
+        return '`' + ri + '`'
 
 
 class SyntaxErrorListener(ErrorListener):
@@ -43,7 +59,6 @@ class ADQLtoMySQLGeometryTranslationVisitor(ADQLParserVisitor):
     """
     def __init__(self, conunits="RADIANS"):
         self.contexts = {}
-        self.limit = None
         self.conunits = conunits
 
     def _convert_values(self, ctx, cidx):
@@ -77,32 +92,24 @@ class ADQLtoMySQLGeometryTranslationVisitor(ADQLParserVisitor):
         return vals
 
     def visitRegular_identifier(self, ctx):
-        ri = ctx.getText().rstrip("'").lstrip("'").rstrip('"').lstrip('"')
-        ri = '`' + ri + '`'
+        if isinstance(ctx.parentCtx,
+                      ADQLParser.User_defined_function_nameContext):
+            return
+        ri = _process_regular_identifier(ctx.getText())
         self.contexts[ctx] = ri
 
     def visitSchema_name(self, ctx):
-        ri = ctx.getText().rstrip("'").lstrip("'").rstrip('"').lstrip('"')
-        ri = '`' + ri + '`'
+        ri = _process_regular_identifier(ctx.getText())
         self.contexts[ctx] = ri
 
     def visitAs_clause(self, ctx):
         # We need to visit the AS clause to avoid aliases being treated same
         # as regular identifiers and backticked.
-        pass
-
-    def visitSet_limit(self, ctx):
-        """
-        TOP N needs to go to the back of the query and it needs to become
-        LIMIT N.
-
-        :param ctx:
-            antlr context.
-
-        """
-        self.limit = int(ctx.children[1].getText())
-        ctx.removeLastChild()
-        ctx.removeLastChild()
+        ri = _process_regular_identifier(ctx.children[1].getText())
+        if ctx.children[1].getText()[0] != '"':
+            ri = ri.replace('`', '')
+        _remove_children(ctx)
+        self.contexts[ctx] = 'AS ' + ri
 
     def visitPoint(self, ctx):
         coords = []
@@ -228,8 +235,59 @@ class ADQLtoMySQLFunctionsTranslationVisitor(ADQLParserVisitor):
         _remove_children(ctx)
         self.contexts[ctx] = ctx_text
 
-    def visitTable_reference(self, ctx):
-        print(ctx.getText())
+
+class SelectQueryListener(ADQLParserListener):
+    def __init__(self):
+        self.limit_visitor = LimitVisitor()
+        self.limit_remover = LimitVisitor(remove=True)
+        self.limit_terminal_visitor = LimitTerminalVisitor()
+        self.select_query_count = 1
+        self.limit_contexts = {}
+
+    def enterSelect_query(self, ctx):
+        self.limit_visitor.reset_limit_ctx()
+        self.limit_visitor.visit(ctx)
+        self.select_query_count = len(self.limit_visitor.limit_ctx)
+
+        if len(self.limit_visitor.limit_ctx) == 1:
+
+            self.limit_remover.visit(ctx)
+            self.limit_terminal_visitor.visit(ctx)
+            lstr = 'LIMIT %d' % self.limit_remover.limit
+            self.limit_contexts[self.limit_terminal_visitor.terminal] = lstr
+
+
+class LimitVisitor(ADQLParserVisitor):
+    def __init__(self, remove=False):
+        self.reset_limit_ctx()
+        self.remove = remove
+        self.limit = None
+
+    def reset_limit_ctx(self):
+        self.limit_ctx = []
+
+    def visitSet_limit(self, ctx):
+        if self.remove:
+            try:
+                self.limit = int(ctx.children[1].getText())
+                ctx.removeLastChild()
+                ctx.removeLastChild()
+            except IndexError:
+                pass
+        else:
+            try:
+                ctx.children[1]
+                self.limit_ctx.append(ctx)
+            except IndexError:
+                pass
+
+
+class LimitTerminalVisitor(ADQLParserVisitor):
+    def __init__(self):
+        self.terminal = None
+
+    def visitTerminal(self, ctx):
+        self.terminal = ctx
 
 
 class FormatListener(ADQLParserListener):
@@ -237,28 +295,42 @@ class FormatListener(ADQLParserListener):
     Used for formating the output query.
 
     """
-    def __init__(self, parser, contexts, limit):
+    def __init__(self, parser, contexts, limit_contexts):
         self._parser = parser
         self.nodes = []
         self.contexts = contexts
-        self.limit = limit
+        self.limit_contexts = limit_contexts
 
     def visitTerminal(self, node):
         try:
-            self.nodes.append(self.contexts[node.parentCtx])
+            nd = self.contexts[node.parentCtx]
         except KeyError:
-            self.nodes.append(node.getText())
+            nd = node.getText()
+
+        if isinstance(node.parentCtx, ADQLParser.Set_function_typeContext)\
+            or isinstance(node.parentCtx.parentCtx,
+                          ADQLParser.User_defined_function_nameContext)\
+                or nd.upper() in adql_function_names:
+            nd += '_'
+
+        self.nodes.append(nd)
+
+        try:
+            nd = self.limit_contexts[node]
+            self.nodes.append(nd)
+        except KeyError:
+            pass
 
     def format_query(self):
-        query = ' '.join(self.nodes)
+        query = ' '.join(self.nodes).rstrip(';')
+        query = query.replace('_ ', '')
         # Remove some spaces
         query = query.replace(' . ', '.')
         query = query.replace(' , ', ', ')
         query = query.replace('( ', '(')
         query = query.replace(' )', ')')
-        if self.limit:
-            query += ' LIMIT %d' % self.limit
-        return query
+        query = query.rstrip()
+        return '%s;' % query.rstrip()
 
 
 class ADQLQueryTranslator(object):
@@ -271,7 +343,6 @@ class ADQLQueryTranslator(object):
     """
     def __init__(self, query=None):
         self.syntax_error_listener = SyntaxErrorListener()
-        self.parsed = False
         self._query = None
 
         if query is not None:
@@ -288,12 +359,13 @@ class ADQLQueryTranslator(object):
         self.parser = ADQLParser(self.stream)
         self.parser._listeners = [self.syntax_error_listener]
 
-        self.tree = self.parser.query_specification()
-        self.parsed = True
+        self.tree = self.parser.query()
+        #  self.parsed = True
+        self.syntax_errors = self.syntax_error_listener.syntax_errors
 
         if len(self.syntax_error_listener.syntax_errors):
-            print(self.syntax_error_listener.syntax_errors)
-            raise RuntimeError("ADQL query has errors.")
+            raise QuerySyntaxError(self.syntax_error_listener.syntax_errors)
+
         self.walker = antlr4.ParseTreeWalker()
 
     @property
@@ -312,7 +384,7 @@ class ADQLQueryTranslator(object):
             Query string.
 
         """
-        self._query = query.rstrip(';') + ';'
+        self._query = query.lstrip('\n').rstrip().rstrip(';') + ';'
         self.parse()
 
     def to_mysql(self):
@@ -321,15 +393,24 @@ class ADQLQueryTranslator(object):
         for the spherical functions.
 
         """
-        assert self.parsed, 'No query given or query not parsed yet.'
+        if self._query is None:
+            raise QueryError('No query given.')
+
+        self.parse()
+
         translator_visitor = ADQLtoMySQLGeometryTranslationVisitor()
         translator_visitor.visit(self.tree)
-        limit = translator_visitor.limit
         translator_visitor = \
             ADQLtoMySQLFunctionsTranslationVisitor(translator_visitor.contexts)
         translator_visitor.visit(self.tree)
-        self.format_listener = FormatListener(self.parser,
-                                              translator_visitor.contexts,
-                                              limit)
-        self.walker.walk(self.format_listener, self.tree)
-        return self.format_listener.format_query()
+
+        select_query_listener = SelectQueryListener()
+        while select_query_listener.select_query_count > 0:
+            self.walker.walk(select_query_listener, self.tree)
+        self.walker.walk(select_query_listener, self.tree)
+
+        format_listener = FormatListener(self.parser,
+                                         translator_visitor.contexts,
+                                         select_query_listener.limit_contexts)
+        self.walker.walk(format_listener, self.tree)
+        return format_listener.format_query()
